(    Title:  Tests for modules.fs
    Author:  David N. Williams
      File:  modules-test.fs
   License:  Public Domain
   Version:  0.5.2
   Revised:  February 16, 2012

The date above may reflect cosmetic changes not logged here.

Version 0.5.2
15Feb12 * Added test for MODULES-VERSION.
16Feb12 * Revised tests invoving MEMBER-FIND to compare EXECUTE
	  results instead of xt's, to accomodate iForth's dual
	  xt system.

Version 0.5.0
15Feb12 * Added tests for new exceptions in:
	  PRIVATE: PUBLIC: [THIS]

Version 0.4.4
13Feb12 * Added SHOW-MODULES test.
14Feb12 * Added tests for documentation Foo Bar example.
	* Added MODULE-DO-REF tests.
	* Added tests for clearing of PUBLIC-DEFS and
	  PRIVATE-DEFS.
	* Fixed bug in max ostack size report.

Version 0.4.0
11Feb12 * Started, based on root-modules-test.fs 0.8.3.
12Feb12 * First release.

This file tests the modules library by Krishna Myneni and
David N. Williams.  For completeness, it copies tests from
order-test.fs 1.3.1 for the search-order utility words in
order.fs 1.3.1 that are also defined in modules.fs.

This file contains some non-ASCII UTF-8 characters.
)
decimal

\ for debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]
: .#       ( n -- )  cr ." #" . cr ;

s" ttester-xf.fs"     INCLUDED
s" otester.fs"        INCLUDED
s" tester-display.fs" INCLUDED
s" xtester-errors.fs" INCLUDED
s" otester-errors.fs" INCLUDED

true VERBOSE !

\ before first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

: ?.errors  ( -- )
  VERBOSE @
  IF blue-text ." XT-#ERRORS: " normal-text XT-#ERRORS @ .
     3 spaces
     blue-text ." OT-#ERRORS: " normal-text OT-#ERRORS @ .
  THEN ;

: ={  O{ T{ ;    : =>   O-> -> ;    : }=  }O }T ;

\ for debugging
: .#order  ( n -- )  cr 2 .r ." : " get-order .s drops ;

PFE-HOST [IF]    ?." Host is pfe, "
[THEN]
GFORTH-HOST [IF] ?." Host is gforth, "
[THEN]
IFORTH-HOST [IF] ?." Host is iForth, "
[THEN]
?." Maximum size of search order: " OT-MAX#OSTACK ?. ?.cr


\ *** ORDER STACK UTILITIES
(
The words >O, O>, and O@ below are versions of >ORDER, ORDER>,
and ORDER@ that have been copied here from already tested
definitions in order.fs.
)

: >O  ( wid -- ) ( o: -- wid )
(
Move the wid from the data stack to the order stack.
)
  >r get-order r> swap 1+ set-order ;

: O>  ( o: wid -- ) ( -- wid )
(
Move the wid from the order stack to the data stack. 
)
  get-order swap >r 1- set-order r> ;

: O@  ( o: wid -- ) ( -- wid )
  get-current definitions get-current swap set-current ;


\ *** TESTS

TESTING  state after loading modules.fs

get-current CONSTANT env-current
         o@ CONSTANT env-context

\ this test does not leave MODULES-WORDLIST in the search order
={ s" modules.fs" INCLUDED => MODULES-WORDLIST >o }=
t{ get-current o@ -> env-current env-context }t

?." Modules version is " MODULES-VERSION ?type ?." .

\ The system might clear Forth variables, so inspect the source.
t{ public-defs @ private-defs @ -> 0 0 }t

TESTING  >ORDER ORDER> ORDER@ ORDER-DEPTH ORDER-DROPS DROPS

wordlist constant wid1
wordlist constant wid2
wordlist constant wid3

={ wid1 >order previous => }=

={ get-order wid1 >order order> => get-order wid1 }=
={ get-order wid1               => get-order wid1 >order order> }=

={ wid1 >order get-order order> >order
     => wid1 >order get-order }=

: >o:  ( wid <"name"> -- )  create ,
DOES>  ( o: -- wid )        @ >o ;

wid1 >o: wid1>o
wid2 >o: wid2>o
wid3 >o: wid3>o

={ wid1>o => wid1 >order }=
={ wid2>o => wid2 >order }=
={ wid3>o => wid3 >order }=

={ wid1>o order@ previous => wid1 }=

={ get-order => get-order drop order-depth }=

={ 0 order-drops => }=
={ wid1>o 1 order-drops => }=
={ wid1>o wid2>o wid3>o 2 order-drops => wid1>o }=

={ 0 drops => }=
={ 2 1 drops => }=
={ 1 2 3 2 drops => 1 }=

TESTING  SAVE-SO-STATE RESTORE-SO-STATE

t{ order-depth save-SO-state env-current
     -> initial-order-depth @ initial-defs @ }t
={ restore-SO-state => }=
={ wid1>o wid2>o save-SO-state restore-SO-state get-current
     => wid1>o wid2>o env-current }=

TESTING  NAMED-MODULE MODULES MODULE: BEGIN-MODULE PRIVATE: PUBLIC:

={ named-module @ => 0 }=
={ ALSO MODULES => MODULES-WORDLIST >order }=

={ BEGIN-MODULE get-current env-current set-current
    => env-current >order private-defs @ >order Private-Words }=

\ Later there'll be a module Foo that uses Bar.
={ MODULE: Bar get-current => env-current }=

={ ALSO MODULES ' Bar >body order> drop => named-module @ }=

={ BEGIN-MODULE get-current
     => named-module @ dup @ ( public_wid)  >order
                     cell+ @ ( private_wid) >order order@ }=

env-current set-current
named-module @ dup @ VALUE public-wid
             cell+ @ VALUE private-wid
t{ public-wid private-wid <> -> true }t

={ PRIVATE: get-current => private-wid }=
={ PUBLIC:  get-current => public-wid }=
env-current set-current

={ ALSO MODULES Bar order> => public-wid }=

: 0vars  0 public-defs ! 0 private-defs ! ;

0vars
\ Whether a message is displayed depends on the implementation
\ of CATCH.
VERBOSE @ [IF]
?." Your system may display two, 'not in module' messages:
t{ ' PUBLIC:  CATCH -> -1 }t
t{ ' PRIVATE: CATCH -> -1 }t
[THEN]
\ Uncomment one of the following to see the exception:
\ PUBLIC:
\ PRIVATE:

TESTING  >PUBLIC >PRIVATE END-MODULE

ALSO MODULES
={ ' Bar >public ' Bar >private  => public-wid private-wid }=
previous

\ set state just after BEGIN-MODULE:
t{ private-wid set-current public-wid >order private-wid >order -> }t

\ drops wids
={ END-MODULE public-wid >order private-wid >order => }=

\ restores current, clears NAMED-MODULE
t{ get-current named-module @ -> env-current 0 }t

\ back to original state
t{ END-MODULE get-current order@ -> env-current env-context }t
t{ public-defs @ private-defs @ -> 0 0 }t

TESTING  MEMBER-FIND MEMBER-NOT-FOUND MODULE-DO-REF

order-depth VALUE initial-odepth

PFE-HOST 0= [IF]  \ pfe fails
={ member-find FORTH + => ' + -1 }=
[THEN]

ALSO MODULES ALSO Bar
public-wid public-defs !    private-wid private-defs !

PUBLIC:
: trece  13 ;   : itrece 13 ; immediate

\ for future use by Foo
: Get  ( -- s )  s" This is the Bar version of 'Get'" ;

={ member-find Bar itrece swap execute => 1 ' itrece execute }=
={ member-find Bar catorce => 0 }=
VERBOSE @ [IF]
?." Your system may display a 'not found' message for 'catorce':
={ ' member-not-found CATCH => -1 }=
\ Uncomment the following to see the exception:
\ member-not-found
[THEN]

={ member-find Bar trece swap execute  => -1 ' trece execute }=

={ ' trece  1 module-do-ref => 13 }=
={ ' trece -1 module-do-ref => 13 }=
={ : compiled-trece  [ ' trece ] literal -1 module-do-ref ;
     compiled-trece => 13 }=

PREVIOUS PREVIOUS
0vars
t{ order-depth -> initial-odepth }t

?." The next TESTING line may display the 'contains as member' and
?." 'there exists' characters incorrectly if your terminal is not
?." set for UTF-8:
TESTING  M' [M'] [M] ∋ [THIS] ∃
(
Here we make a new module Foo that defines words with the same
names as those in Bar.  Testing takes place within Foo.
)

order-depth to initial-odepth

ALSO MODULES
MODULE: Foo
ALSO Bar
BEGIN-MODULE

={ m' Bar itrece execute => ' itrece execute }=
={ m' Bar trece execute  => ' trece execute }=

VERBOSE @ [IF] IFORTH-HOST 0= [IF]
\ iForth tries to interepret CATORCE and throws us out
?." Your system may display a 'not found' message for 'catorce':
={ ' m' CATCH Bar catorce => -1 }=
[THEN] [THEN]
\ Uncomment the following line to see the exception:
\ m' Bar catorce

={ : itrece [m'] Bar itrece ; itrece => m' Bar itrece }=
\ iForth fails:
={ : trece  [m'] Bar trece ;  trece  => m' Bar trece }=

={ [m] Bar trece => 13 }=
={ ∋   Bar trece => 13 }=

\ Uncomment the following to see the exception:
\ : catorce  [m'] Bar catorce ;

PUBLIC:
: quince  15 ;
PRIVATE:
: dieciséis  16 ;

={ [this] quince => 15 }=
={      ∃ quince => 15 }=
={ [this] dieciséis => 16 }=
={      ∃ dieciséis => 16 }=

\ Uncomment one of the following lines to see an exception:
\ [this] +
\ ∃ +

PUBLIC:  \ test for documentation example

: Insert1  ( -- s )
  [ ALSO Bar ] Get [ PREVIOUS ] ;

: Insert2  ( -- s)
  ∋ Bar Get ;

={ Insert1 Get compare => 0 }=
={ Insert2 Get compare => 0 }=

END-MODULE
PREVIOUS

VERBOSE @ [IF] IFORTH-HOST 0= [IF]
\ iForth tries to interpret QUINCE and throws us out.
?." Your system may display a 'not in body' message.
={ ' [this] CATCH quince => -1 }=
[THEN] [THEN]
\ Uncommenet the following line to see the exception:
\ [this] quince

t{ order-depth -> initial-odepth }t

TESTING  [MEMBER]

[member] Forth scan [IF]
?." SCAN is a member of FORTH
[ELSE]
?." SCAN is not a member of FORTH
[THEN]

[member] Forth + [IF]
?." + is a member of FORTH
[ELSE]
?." + is not a member of FORTH
[THEN]

VERBOSE @ [IF]
TESTING  SHOW-MODULES

?." You should see the test module names 'Foo' and 'Bar':"
show-modules IFORTH-HOST 0= [IF] ?.cr [THEN]
[THEN]

?.errors GFORTH-HOST [IF] ?.cr [THEN]
